package edu.sdsc.inca.depot.persistent;

import java.math.BigInteger;
import org.apache.xmlbeans.XmlObject;

/**
 * Schedule holds the information for a cron-like schedule that we use.  In the
 * future there will be other schedule types as well.
 *
 * @author cmills
 */
public class Schedule extends PersistentObject {

  private String minute;
  private String hour;
  private String month;
  private String mday;
  private String wday;
  private String type;
  private Integer numOccurs;
  private boolean suspended;

  /**
   * Default constructor.
   */
  public Schedule() {
    this("*", "*", "*", "*", "*", "cron", new Integer(-1));
  }

  /**
   * Full constructor.
   *
   * @param min
   * @param hour
   * @param month
   * @param mday
   * @param wday
   * @param type
   * @param numOccurs
   */
  public Schedule(String min, String hour, String month, String mday,
                  String wday, String type, Integer numOccurs) {
    this.setMinute(min);
    this.setHour(hour);
    this.setMonth(month);
    this.setMday(mday);
    this.setWday(wday);
    this.setType(type);
    this.setNumOccurs(numOccurs);
    this.setSuspended(false);
  }

  /**
   * Copies information from an Inca schema XmlBean Schedule object so that
   * this object contains equivalent information.
   *
   * @param o the XmlBean Schedule object to copy
   * @return this, for convenience
   */
  public PersistentObject fromBean(XmlObject o) {
    return this.fromBean((edu.sdsc.inca.dataModel.util.Schedule)o);
  }

  /**
   * Copies information from an Inca schema XmlBean Schedule object so that
   * this object contains equivalent information.
   *
   * @param s the XmlBean Schedule object to copy
   * @return this, for convenience
   */
  public Schedule fromBean(edu.sdsc.inca.dataModel.util.Schedule s) {
    edu.sdsc.inca.dataModel.util.Cron cron = s.getCron();
    if(cron == null) {
      this.setType("immediate");
    } else {
      this.setType("cron");
      this.setMday(cron.getMday());
      this.setHour(cron.getHour());
      this.setMinute(cron.getMin());
      this.setMonth(cron.getMonth());
      this.setWday(cron.getWday());
      BigInteger numOccurs = cron.getNumOccurs();
      if(numOccurs != null) {
        this.setNumOccurs(new Integer(numOccurs.intValue()));
      }
      this.setSuspended(cron.getSuspended());
    }
    return this;
  }

  /**
   * Return the minute portion of the cron schedule.
   *
   * @return the minute cron field
   */
  public String getMinute() {
    return this.minute;
  }

  /**
   * Set the minute portion of the cron schedule.
   *
   * @param minute this minute portion of the schedule
   */
  public void setMinute(String minute) {
    if(minute == null || minute.equals("")) {
      minute = DB_EMPTY_STRING;
    }
    this.minute = truncate(minute, MAX_DB_STRING_LENGTH, "minute");
  }

  /**
   * Return the hour portion of the cron schedule.
   *
   * @return the hour cron field
   */
  public String getHour() {
    return this.hour;
  }

  /**
   * Set the hour portion of the cron schedule.
   *
   * @param hour this hour portion of the schedule
   */
  public void setHour(String hour) {
    if(hour == null || hour.equals("")) {
      hour = DB_EMPTY_STRING;
    }
    this.hour = truncate(hour, MAX_DB_STRING_LENGTH, "hour");
  }

  /**
   * Return the month portion of the cron schedule.
   *
   * @return the month cron field
   */
  public String getMonth() {
    return this.month;
  }

  /**
   * Set the month portion of the cron schedule.
   *
   * @param month the month portion of the schedule
   */
  public void setMonth(String month) {
    if(month == null || month.equals("")) {
      month = DB_EMPTY_STRING;
    }
    this.month = truncate(month, MAX_DB_STRING_LENGTH, "month");
  }

  /**
   * Return the mday portion of the cron schedule.
   *
   * @return the month day cron field
   */
  public String getMday() {
    return this.mday;
  }

  /**
   * Set the month day portion of the schedule.
   *
   * @param mday the month day portion of the schedule
   */
  public void setMday(String mday) {
    if(mday == null || mday.equals("")) {
      mday = DB_EMPTY_STRING;
    }
    this.mday = truncate(mday, MAX_DB_STRING_LENGTH, "mday");
  }

  /**
   * Return the wday portion of the cron schedule.
   *
   * @return the week day cron field.
   */
  public String getWday() {
    return this.wday;
  }

  /**
   * Set the week day portion of the schedule.
   *
   * @param wday the week day portion of the schedule
   */
  public void setWday(String wday) {
    if(wday == null || wday.equals("")) {
      wday = DB_EMPTY_STRING;
    }
    this.wday = truncate(wday, MAX_DB_STRING_LENGTH, "wday");
  }

  /**
   * Returns the type of schedule.
   *
   * @return the type of schedule
   */
  public String getType() {
    return this.type;
  }

  /**
   * Set the type of schedule.  Current recognized values are "cron" and
   * "immediate".
   *
   * @param type
   */
  public void setType(String type) {
    if(type == null || type.equals("")) {
      type = DB_EMPTY_STRING;
    }
    this.type = truncate(type, MAX_DB_STRING_LENGTH, "schedule type");
  }

  /**
   * Null implies unlimited.
   *
   * @return the number of times the schedule will be executed.
   */
  public Integer getNumOccurs() {
    return numOccurs;
  }

  /**
   * Set the number of times the schedule should be run before being
   * discarded.  Null indicates unlimited.
   * @param numOccurs
   */
  public void setNumOccurs(Integer numOccurs) {
    this.numOccurs = numOccurs;
  }

  /**
   * Returns whether or not further execution has been suspended pending
   * a user request to resume.
   * @return the suspension status
   */
  public boolean getSuspended() {
    return this.suspended;
  }

  /**
   * Indicates whether or not further execution should be suspended pending
   * a user request to resume.
   * @param suspended the suspension status
   */
  public void setSuspended(boolean suspended) {
    this.suspended = suspended;
  }

  /**
   * Returns a Inca schema XmlBean Schedule object that contains information
   * equivalent to this object.
   *
   * @return an XmlBean Schedule object that contains equivalent information
   */
  public XmlObject toBean() {
    edu.sdsc.inca.dataModel.util.Schedule result =
      edu.sdsc.inca.dataModel.util.Schedule.Factory.newInstance();
    if(this.getType().equals("cron")) {
      edu.sdsc.inca.dataModel.util.Cron cron = result.addNewCron();
      cron.setMday(this.getMday());
      cron.setHour(this.getHour());
      cron.setMin(this.getMinute());
      cron.setMonth(this.getMonth());
      cron.setWday(this.getWday());
      cron.setNumOccurs(BigInteger.valueOf(
        this.getNumOccurs() == null ? -1 : this.getNumOccurs().intValue()
      ));
      cron.setSuspended(this.getSuspended());
    }
    return result;
  }

  public String toString() {
    return this.getMinute() + "," +
           this.getHour() + "," +
           this.getMonth() + "," +
           this.getMday() + "," +
           this.getWday() + "," +
           this.getType() + "," +
           this.getNumOccurs() + "," +
           new Boolean(this.getSuspended());
  }

  /**
   * Override of the default equals method.
   */
  public boolean equals(Object o) {
    return this.toString().equals(o.toString());
  }

}
